@@ -4,6 +4,7 @@ use 5.020;
 use strict;
 use warnings;
 
+use Time::HiRes ();
 use Types::Common -types;
 use Marlin
     'prompt!',                                     # Required prompt (string or async generator)
@@ -27,7 +28,7 @@ use Marlin
                 # with streaming JSON and partial data
                 warn "JSON::Lines $action error: $error"
                     if $ENV{CLAUDE_AGENT_DEBUG} && $ENV{CLAUDE_AGENT_DEBUG} > 1;
-                return undef;
+                return;
             },
         )
     };
@@ -39,6 +40,7 @@ use Future::AsyncAwait;
 use JSON::Lines;
 use Try::Tiny;
 use File::Which qw(which);
+use File::Spec;
 
 use Claude::Agent::Options;
 use Claude::Agent::Message;
@@ -126,6 +128,7 @@ sub BUILD {
     }
 
     $self->_start_process();
+    return;
 }
 
 sub _find_claude_cli {
@@ -139,9 +142,17 @@ sub _find_claude_cli {
     my @paths = (
         '/usr/local/bin/claude',
         '/opt/homebrew/bin/claude',
-        "$ENV{HOME}/.local/bin/claude",
-        "$ENV{HOME}/.npm-global/bin/claude",
     );
+    # Add HOME-based paths only if HOME is a valid absolute path without traversal
+    if ($ENV{HOME} && $ENV{HOME} =~ m{^/}) {
+        require Cwd;
+        my $home = Cwd::abs_path($ENV{HOME});
+        # Only use if resolved path is valid and absolute
+        if ($home && $home =~ m{^/} && -d $home) {
+            push @paths, File::Spec->catfile($home, '.local', 'bin', 'claude');
+            push @paths, File::Spec->catfile($home, '.npm-global', 'bin', 'claude');
+        }
+    }
 
     for my $path (@paths) {
         return $path if -x $path;
@@ -150,6 +161,7 @@ sub _find_claude_cli {
     Claude::Agent::Error::CLINotFound->throw(
         message => "Could not find 'claude' CLI in PATH or common locations"
     );
+    return;  # Never reached, but satisfies perlcritic
 }
 
 sub _build_command {
@@ -167,9 +179,9 @@ sub _build_command {
         push @cmd, '--model', $opts->model;
     }
 
-    # Add max turns if specified
+    # Add max turns if specified (ensure integer)
     if ($opts->has_max_turns && $opts->max_turns) {
-        push @cmd, '--max-turns', $opts->max_turns;
+        push @cmd, '--max-turns', int($opts->max_turns);
     }
 
     # Add permission mode
@@ -201,16 +213,22 @@ sub _build_command {
     if ($opts->has_system_prompt && $opts->system_prompt) {
         my $sp = $opts->system_prompt;
         if (ref $sp eq 'HASH' && $sp->{preset}) {
-            push @cmd, '--system-prompt', $sp->{preset};
+            # Sanitize control characters from preset name
+            my $sanitized = $sp->{preset};
+            $sanitized =~ s/[[:cntrl:]]/ /g;
+            push @cmd, '--system-prompt', $sanitized;
         }
         elsif (!ref $sp) {
-            push @cmd, '--system-prompt', $sp;
+            # Sanitize control characters from system prompt
+            my $sanitized = $sp;
+            $sanitized =~ s/[[:cntrl:]]/ /g;
+            push @cmd, '--system-prompt', $sanitized;
         }
     }
 
     # Add MCP servers config
     # SDK servers are converted to stdio servers pointing to our SDKRunner
-    if ($opts->has_mcp_servers && $opts->mcp_servers || keys %{$self->_sdk_servers}) {
+    if (($opts->has_mcp_servers && $opts->mcp_servers && keys %{$opts->mcp_servers}) || keys %{$self->_sdk_servers}) {
         my %servers;
 
         # Add non-SDK servers directly
@@ -268,7 +286,15 @@ sub _build_command {
     # For string prompts, use --print mode with -- separator
     # For async generators, use stream-json input format
     if (!ref($self->prompt)) {
-        push @cmd, '--print', '--', $self->prompt;
+        # Sanitize control characters from prompt to prevent injection
+        # Preserve tabs (\t), newlines (\n, \r) which are common in code snippets
+        my $sanitized_prompt = $self->prompt;
+        $sanitized_prompt =~ s/\x1b\[[0-9;]*[a-zA-Z]//g;  # Strip ANSI escape codes
+        $sanitized_prompt =~ s/\x1b\][^\x07]*\x07//g;     # Strip OSC sequences (title changes, etc.)
+        $sanitized_prompt =~ s/\x1b[PX^_].*?\x1b\\//g;    # Strip DCS/SOS/PM/APC sequences
+        # Remove dangerous control chars but preserve tab (\x09), newline (\x0a), carriage return (\x0d)
+        $sanitized_prompt =~ s/[\x00-\x08\x0b\x0c\x0e-\x1f\x7f]/ /g;
+        push @cmd, '--print', '--', $sanitized_prompt;
     }
     else {
         # For streaming input (async generator), use stream-json input format
@@ -283,7 +309,11 @@ sub _start_process {
 
     my @cmd = $self->_build_command();
 
-    warn "DEBUG: Running command: @cmd\n" if $ENV{CLAUDE_AGENT_DEBUG};
+    if ($ENV{CLAUDE_AGENT_DEBUG}) {
+        my @safe_cmd = map { $_ =~ /^--/ ? $_ : '[REDACTED]' } @cmd[0..2];
+        push @safe_cmd, '...' if @cmd > 3;
+        warn "DEBUG: Running command: @safe_cmd (full args redacted)\n";
+    }
 
     my $process = IO::Async::Process->new(
         command => \@cmd,
@@ -291,7 +321,7 @@ sub _start_process {
         stdout => {
             on_read => sub {
                 my ($stream, $buffref) = @_;
-                while ($$buffref =~ s/^(.+)\n//) {
+                while ($$buffref =~ s/^([^\n]+)\n//) {
                     my $line = $1;
                     $self->_handle_line($line);
                 }
@@ -302,7 +332,7 @@ sub _start_process {
             on_read => sub {
                 my ($stream, $buffref) = @_;
                 # Log stderr but don't treat as fatal
-                while ($$buffref =~ s/^(.+)\n//) {
+                while ($$buffref =~ s/^([^\n]+)\n//) {
                     warn "Claude CLI stderr: $1\n" if $ENV{CLAUDE_AGENT_DEBUG};
                 }
                 return 0;
@@ -311,8 +341,10 @@ sub _start_process {
         on_finish => sub {
             my ($proc, $exitcode) = @_;
             $self->_finished(1);
-            if ($exitcode != 0) {
-                $self->_error("Claude CLI exited with code $exitcode");
+            # Extract actual exit status (WEXITSTATUS equivalent)
+            my $exit_status = $exitcode >> 8;
+            if ($exit_status != 0) {
+                $self->_error("Claude CLI exited with code $exit_status");
             }
             # Resolve any pending async futures
             $self->_resolve_pending_futures_on_finish();
@@ -326,15 +358,18 @@ sub _start_process {
         },
     );
 
-    $self->_loop->add($process);
+    # Store references before adding to loop to avoid race conditions
     $self->_process($process);
     $self->_stdin($process->stdin);
+    $self->_loop->add($process);
 
-    # For non-streaming (--print) mode, close stdin to signal we're done
+    # For --print mode, prompt is in argv, stdin can be closed
     if (!ref($self->prompt)) {
-        $self->_stdin->close_when_empty;
+        # Schedule close after a brief delay to allow process startup
+        $self->_loop->later(sub { $self->_stdin->close_when_empty if $self->_stdin });
     }
     # For ref prompts (streaming input), caller will send messages via send_user_message
+    return;
 }
 
 sub _handle_line {
@@ -353,6 +388,33 @@ sub _handle_line {
             warn "DEBUG: Buffer content (first 200): " . substr($self->_jsonl->remaining, 0, 200) . "\n";
         }
     }
+
+    # Guard against buffer overflow from accumulated malformed data
+    # Configurable threshold via environment variable, default 100KB, max 10MB
+    my $buffer_threshold = $ENV{CLAUDE_AGENT_JSONL_BUFFER_MAX} // 100_000;
+    my $min_threshold = 1000;  # Minimum 1KB to prevent excessive reinitializations
+    if (!defined $buffer_threshold || $buffer_threshold !~ /^\d+$/ || $buffer_threshold < $min_threshold) {
+        warn "Invalid CLAUDE_AGENT_JSONL_BUFFER_MAX value, using default 100000\n"
+            if defined $ENV{CLAUDE_AGENT_JSONL_BUFFER_MAX};
+        $buffer_threshold = 100_000;
+    }
+    $buffer_threshold = 10_000_000 if $buffer_threshold > 10_000_000;  # Hard cap at 10MB
+
+    # Check buffer size regardless of decode success to prevent unbounded growth
+    if ($self->_jsonl->remaining && length($self->_jsonl->remaining) > $buffer_threshold) {
+        warn "JSON::Lines buffer overflow detected (size: " . length($self->_jsonl->remaining) .
+             ", threshold: $buffer_threshold), reinitializing\n" if $ENV{CLAUDE_AGENT_DEBUG};
+        $self->_jsonl(JSON::Lines->new(
+            utf8     => 1,
+            error_cb => sub {
+                my ($action, $error, $data) = @_;
+                warn "JSON::Lines $action error: $error"
+                    if $ENV{CLAUDE_AGENT_DEBUG} && $ENV{CLAUDE_AGENT_DEBUG} > 1;
+                return;
+            },
+        ));
+    }
+
     return unless @decoded;
 
     for my $data (@decoded) {
@@ -386,6 +448,7 @@ sub _handle_line {
             push @{$self->_messages}, $msg;
         }
     }
+    return;
 }
 
 sub _resolve_pending_futures_on_finish {
@@ -397,8 +460,13 @@ sub _resolve_pending_futures_on_finish {
 
     # Stop SDK servers
     for my $sdk_server (values %{$self->_sdk_servers}) {
-        $sdk_server->stop();
+        try {
+            $sdk_server->stop();
+        } catch {
+            warn "Failed to stop SDK server: $_" if $ENV{CLAUDE_AGENT_DEBUG};
+        };
     }
+    return;
 }
 
 =head2 next
@@ -409,6 +477,7 @@ Blocking call to get the next message. Returns undef when no more messages.
 
 =cut
 
+## no critic (ProhibitBuiltinHomonyms)
 sub next {
     my ($self) = @_;
 
@@ -416,8 +485,33 @@ sub next {
     return shift @{$self->_messages} if @{$self->_messages};
 
     # Wait for more messages or process to finish
-    while (!@{$self->_messages} && !$self->_finished) {
-        $self->_loop->loop_once(0.01);
+    # Configurable timeout with 10 minute default, max 1 hour
+    my $has_timeout = $self->options->has_query_timeout;
+    my $timeout = $self->options->query_timeout // 600;
+    my $max_timeout = 3600;  # 1 hour maximum
+    if (!defined $timeout || $timeout <= 0 || $timeout > $max_timeout) {
+        warn "Invalid query_timeout value ($timeout), using default 600 seconds\n"
+            if $has_timeout && defined $timeout && $timeout != 600 && $ENV{CLAUDE_AGENT_DEBUG};
+        $timeout = 600;
+    }
+    my $start_time = Time::HiRes::time();
+    # Use epsilon tolerance (0.001s) to handle floating-point boundary conditions
+    my $epsilon = 0.001;
+    while (!@{$self->_messages} && !$self->_finished && !$self->_error
+           && (Time::HiRes::time() - $start_time) < ($timeout - $epsilon)) {
+        $self->_loop->loop_once(0.1);  # Longer interval to reduce CPU busy-waiting
+    }
+
+    # Check if we timed out waiting for messages
+    # Use consistent comparison with epsilon tolerance to avoid boundary race
+    if ((Time::HiRes::time() - $start_time) >= ($timeout - $epsilon)
+        && !$self->_finished) {
+        # Final check for messages that may have arrived at timeout boundary
+        $self->_loop->loop_once(0) if !@{$self->_messages};
+        return shift @{$self->_messages} if @{$self->_messages};
+        $self->_finished(1);
+        $self->_error("Query timed out after $timeout seconds");
+        return;  # Return immediately on timeout
     }
 
     return shift @{$self->_messages};
@@ -501,10 +595,12 @@ Send interrupt signal to abort current operation.
 sub interrupt {
     my ($self) = @_;
 
-    return unless $self->_stdin;
+    return unless $self->_stdin && !$self->_finished;
 
     my $msg = $self->_jsonl->encode([{ type => 'interrupt' }]);
+    return unless defined $msg && length $msg;
     $self->_stdin->write($msg);
+    return;
 }
 
 =head2 send_user_message
@@ -518,7 +614,7 @@ Send a follow-up user message during streaming.
 sub send_user_message {
     my ($self, $content) = @_;
 
-    return unless $self->_stdin;
+    return unless $self->_stdin && !$self->_finished;
 
     my $msg = $self->_jsonl->encode([{
         type    => 'user',
@@ -527,7 +623,13 @@ sub send_user_message {
             content => $content,
         },
     }]);
-    $self->_stdin->write($msg);
+    my $result;
+    try {
+        $result = $self->_stdin->write($msg);
+    } catch {
+        warn "send_user_message write error: $_" if $ENV{CLAUDE_AGENT_DEBUG};
+    };
+    return $result;
 }
 
 =head2 set_permission_mode
@@ -541,13 +643,14 @@ Change permission mode during streaming.
 sub set_permission_mode {
     my ($self, $mode) = @_;
 
-    return unless $self->_stdin;
+    return unless $self->_stdin && !$self->_finished;
 
     my $msg = $self->_jsonl->encode([{
         type            => 'set_permission_mode',
         permission_mode => $mode,
     }]);
     $self->_stdin->write($msg);
+    return;
 }
 
 =head2 respond_to_permission
@@ -564,7 +667,7 @@ Respond to a permission request.
 sub respond_to_permission {
     my ($self, $tool_use_id, $response) = @_;
 
-    return unless $self->_stdin;
+    return unless $self->_stdin && !$self->_finished;
 
     my $msg = $self->_jsonl->encode([{
         type        => 'permission_response',
@@ -572,6 +675,7 @@ sub respond_to_permission {
         response    => $response,
     }]);
     $self->_stdin->write($msg);
+    return;
 }
 
 =head2 rewind_files
@@ -585,10 +689,11 @@ Revert file changes to the checkpoint state.
 sub rewind_files {
     my ($self) = @_;
 
-    return unless $self->_stdin;
+    return unless $self->_stdin && !$self->_finished;
 
     my $msg = $self->_jsonl->encode([{ type => 'rewind_files' }]);
     $self->_stdin->write($msg);
+    return;
 }
 
 1;
